@@ -2,6 +2,7 @@ package TestQless;
 use base qw(Test::Class);
 use Redis;
 use Qless::Client;
+use Time::HiRes;
 
 sub setup : Test(setup) {
 	my $self = shift;
@@ -39,4 +40,27 @@ sub teardown : Test(teardown) {
 	$self->{'redis'}->flushdb();
 }
 
+sub time_freeze {
+	my $self = shift;
+	$self->{'_original_sub'} = \&Time::HiRes::time;
+	$self->{'_time'} = Time::HiRes::time;
+	no warnings qw(redefine);
+	*Time::HiRes::time = sub() {
+		return $self->{'_time'};
+	};
+	use warnings;
+}
+
+sub time_unfreeze {
+	my $self = shift;
+	no warnings qw(redefine);
+	*Time::HiRes::time = $self->{'_original_sub'};
+	use warnings;
+}
+
+sub time_advance {
+	my ($self, $value) = @_;
+	$self->{'_time'} += $value;
+}
+
 1;
